#charset "us-ascii"

/* Copyright (c) 2003 Sren J. Lvborg.  All Rights Reserved. */

/*
 *   t3help.t - A TADS 3 helpfile mechanically generated from
 *   the TADS library code.
 */

/*
 *   There's still a few problems.
 *   First of all, comments and strings are limited to 5K, due to the RegEx bug.
 *
 *   symbols are mistakingly identified within strings.
 *   (E.g., see _default_display_fn or UnknownCharSetException.)
 *
 *   The explicit line break-periods are not removed from comments.
 *.  <-- Like this.
 *   Modified by:	modify Thing 7. -- Number properly.
 *   Fix tads-gen fn-set. Fix includes and charset.
 *   Fix actor.t's Topic class. (isKnown doesn't show up.)
 */

#include <tads.h>

#define TEXT_PLAIN	0x0000	// Just show the text plainly.
#define TEXT_BOLD		0x0001	// Wrap the text in <B> ... </B> tags.
#define TEXT_TAGON 	0x0002	// Text follows which are to be shown in the <A>...</A>, but isn't a part of the keyword.
#define TEXT_KEYWORD	0x0004	// Link the text (<A href="!...">...</A>)
#define TEXT_TELETYPE	0x0008	// Wrap the text in <TT> ... </TT> tags.

property banner, lastpage, curPage, searchKeywords, showFullPage, isShowingFull;
property superclasses, subclasses, modifies, replaces, modifiedby, replacedby, includes;

function UpdateBanner() {
 if (!globals.banner) return;
 bannerClear(globals.banner);
 bannerSay(globals.banner, '<BODY bgcolor=statusbg text=statustext>');
 if (systemInfo(SysInfoInterpClass) == SysInfoIClassHTML)
  bannerSay(globals.banner, '<TABLE width=100% cellspacing=0 cellpadding=0><TR valign=top><TD><B>');
 else bannerSay(globals.banner, '<B>');
 
 bannerSay(globals.banner, globals.curPage.bannerTitle);
 bannerSay(globals.banner, '</B> &nbsp;-&nbsp; ');
 bannerSay(globals.banner, globals.curPage.bannerSubTitle);

 if (systemInfo(SysInfoInterpClass) == SysInfoIClassHTML)
  bannerSay(globals.banner, ' <TD align=right>[ <A plain href=".top">Top</A> ] [ <A plain href=".back">Back</A> ]</TR></TABLE>');

 bannerSay(globals.banner, globals.curPage.getNavigStr());
 bannerSizeToContents(globals.banner);
}

function titleSort(a, b) {
 if (dataType(a) == TypeObject) a = a.title;
 if (dataType(b) == TypeObject) b = b.title;
 if ((a = a.toLower()) > (b = b.toLower())) return 1;
 if (a < b) return -1;
 return 0;
}

property filename, header, replaces, modifies;

class page: object {
 bannerTitle { return title; }
 bannerSubTitle = '';
 subpages = [];
 subtitle = '';
 
 getNavigStr() {
  local parentContLen = parent.subpages.length();
  local prev = (pageindex == 1 ? parentContLen : pageindex - 1);
  local next = (pageindex == parentContLen ? 1 : pageindex + 1);
  local s = '[ <A plain href="!!' + parent.subpages[prev].keywords[1] + '">&lt;&lt; ' + parent.subpages[prev].title + '</A> ]' +
         '&emsp;[ <A plain href="!!' + parent.subpages[next].keywords[1] + '">' + parent.subpages[next].title + ' &gt;&gt;</A> ]';
  if (!globals.isShowingFull) s += '&emsp;[<A plain href=".full"> Full page </A>]';
  return s;
 }
 
 addPage(pg) {
  subpages = subpages.append(pg);
  pg.pageindex = subpages.length();
  pg.parent = self;
 }
 
 partKeyword = nil
 makeCurrent(partKeyword_) {
  if (!partKeyword_ || partKeyword_.length() == 0 ) partKeyword = nil;
  else partKeyword = partKeyword_;
  globals.isShowingFull = true;
  
  globals.lastpage = globals.curPage;
  globals.curPage = self;
  clearScreen();
  "<B><BIG><< title >></BIG></B><< subtitle >>\b";
  display();
  UpdateBanner();
 }

 dispObject(obj) {
  if (dataType(obj) == TypeObject)
   "<A plain href=\"!<< obj.keywords[1] >>\"><FONT color=#008000><< obj.title >></FONT></A>";
  else "<< obj >>";
 }
 dispObjectList(lst) {
  lst = lst.sort(nil, titleSort);
  for (local i = 1; i <= lst.length(); i++) {
   if (i > 1) (i == lst.length() ? " and " : ", ");
   dispObject(lst[i]);
  }
  ".";
 } 
 displayContents(lst, sort, cols) {
  if (sort) lst = lst.sort(nil, titleSort);
 
  local len = lst.length();
  if (systemInfo(SysInfoInterpClass) == SysInfoIClassHTML) {
   local rowcnt = (len + cols - 1)/cols;
   "<TABLE><TR valign=top><TD>";
   local x = 0;
   for (local i = 1; i <= len; i++) {
    x++;
    if (x > rowcnt) { "<TD><TD>"; x = 1; }
    "<A plain href=\"!<< lst[i].keywords[1] >>\"><FONT color=#000080><< lst[i].title >></FONT></A>\n";  
   }
   "</TABLE>";
  } else {
   "<TT>";
   local colwidth = 80/cols - 1;
   for (local i = 1; i <= len; i++) {
    local kw = lst[i].keywords[1];
    "<A plain href=\"!<< kw >>\"><FONT color=#000080><< kw.substr(1, colwidth) >></FONT></A>";
    if (i % cols == 0) "\n";
    else for (local j = kw.length(); j <= colwidth; j++) "&nbsp;";    
   }
   "</TT>";
  }
 }
 displayText() {
  for (local i = 1; i <= text.length(); i++) {
   local x = text[i];
   if (dataType(x) == TypeSString) { "<< x >>"; continue; }
   
   local symbol = text[++i]; if (dataType(symbol) == TypeInt) symbol = globals.keywordIndex[symbol];
   local tagon = (x & TEXT_TAGON) != 0 ? text[++i] : '';
   if ((x & TEXT_BOLD) != 0) "<B>";
   if ((x & TEXT_TELETYPE) != 0) "<TT>";
   if ((x & TEXT_KEYWORD) != 0) "<A plain href=\"!<< symbol >>\"><FONT color=#008000>";
   "<< symbol >><< tagon >>";
   if ((x & TEXT_KEYWORD) != 0) "</FONT></A>";
   if ((x & TEXT_TELETYPE) != 0) "</TT>";
   if ((x & TEXT_BOLD) != 0) "</B>";
  }
 }

#define LeadingList  15 // Number of blocks to show _before_ a matching keyword.
#define TrailingList 35 // Number of blocks to show _after_ a matching keyword.

 displayTextPart() {
  globals.isShowingFull = nil;
  local keywords = partKeyword;
  local textlen = text.length();
  local blocks = new Vector(LeadingList, LeadingList); // A circular ring buffer.
  local blockindex = 1; // Position to insert next block entry.
  local listcnt = 0;
  
  for (local i = 1; i <= textlen; i++) {
   local x = text[i];
   if (dataType(x) == TypeSString) {
    if (listcnt-- > 0) tadsSay(x);
    else { blocks[blockindex++] = x; if (blockindex > LeadingList) blockindex = 1; }
    continue;
   }
   
   local symbol = text[++i]; if (dataType(symbol) == TypeInt) symbol = globals.keywordIndex[symbol];
   local tagon = (x & TEXT_TAGON) != 0 ? text[++i] : '';
   local s = '';
   if ((x & TEXT_BOLD) != 0) s += '<B>';
   if ((x & TEXT_TELETYPE) != 0) s += '<TT>';
   if ((x & TEXT_KEYWORD) != 0) s += '<A plain href="!' + symbol + '"><FONT color=#008000>';
   s += symbol + tagon;
   if ((x & TEXT_KEYWORD) != 0) s += '</FONT></A>';
   if ((x & TEXT_TELETYPE) != 0) s += '</TT>';
   if ((x & TEXT_BOLD) != 0) s += '</B>';

   for (local j = 1; j <= keywords.length(); j++) {
    if (symbol.find(keywords[j])) {
     // It's a match. Dump the ring buffer and set listcnt to list the next 5 blocks.
     if (listcnt <= 0) "\b&emdash; &emdash; &emdash; &emdash; &emdash; &emdash; &emdash;\b";
     for (local k = blockindex; k <= LeadingList; k++) tadsSay(blocks[k]);
     for (local k = 1; k < blockindex; k++) tadsSay(blocks[k]);
     blockindex = 1; blocks.fillValue(nil);
     listcnt = TrailingList;
     break;
    }
   }
   
   if (listcnt-- > 0) tadsSay(s);
   else { blocks[blockindex++] = s; if (blockindex > LeadingList) blockindex = 1; }
  }
  if (listcnt <= 0) "\b&emdash; &emdash; &emdash; &emdash; &emdash; &emdash; &emdash;\b";
 }
}

class defpage: page {
 bannerTitle { return '<A plain href="!' + parent.keywords[1] + '">' + parent.title + '</A>'; }
 bannerSubTitle { return title; }
 display() {
  if (partKeyword && !globals.showFullPage && text.length() > 200) displayTextPart();
  else displayText();
 }
}

class objpage: defpage {
 display() {
  "<TABLE bgcolor=#EEEEEE>";
  if (superclasses.length() != 0)	"<TR valign=top><TD>\nSuperclasses: <TD><< dispObjectList(superclasses) >>";
  if (subclasses.length() != 0)	"<TR valign=top><TD>\nSubclasses: <TD><< dispObjectList(subclasses) >>";
  if (modifiedby.length() != 0)	"<TR valign=top><TD>\nModified by: <TD><< dispObjectList(modifiedby) >>";
  if (replacedby.length() != 0)	"<TR valign=top><TD>\nReplaced by: <TD><< dispObjectList(replacedby) >>";
  if (modifies) "<TR valign=top><TD>\nModifies: <TD><< dispObject(modifies) >>.";
  if (replaces) "<TR valign=top><TD>\nReplaces: <TD><< dispObject(replaces) >>.";
  "</TABLE>\b";
  inherited;
 }
}

class filepage: page {
 bannerSubTitle { return header; }
 title { return filename; }
 subtitle { return (copyright != '') ? (' - ' + copyright) : ''; }
 parent = mainPage;
 
 display() {
  // Display charset, perhaps?
  if (includes.length() != 0) "<TABLE bgcolor=#EEEEEE><TR><TD>Includes <TD><< dispObjectList(includes) >></TABLE>\b";
  "<B>Contents</B>\n";
  displayContents(subpages, true, 3);
  "\b<hr>";
  displayText();
 }
}

property gendate;

mainPage: page {
 // subpages = [] -- This prop is modify'ed by t3help_hlp.t.
 title = 'T3 Help';
 bannerSubTitle { return 'An interactive helpfile (generated on ' + gendate + ')'; }
 display() {
  "<B>Files covered by this help</B>\n";
  displayContents(subpages - instructions, true, 4); 
 }
 getNavigStr = ''; // This is the top-level page.
}

// This includes the actual help texts, which are generated by t3helpgen. (It's big!)
#include "t3help_hlp.h"

PreinitObject {
 execute() {
  for (local i = 1, local len = globals.keywordIndex.length(); i <= len; i++)
   globals.keywordLookup[globals.keywordIndex[i]] = i;
 }
}

modify globals {
 results = [];
 keywordLookup = static new LookupTable(512, 4096);
}

instructions: page {
 title = 'Instructions';
 keywords = ['?']
 bannerSubTitle = 'that is to say, commands';
 display {
  "<B>Commands</B>\n
   <tt>.back&nbsp;</tt> - Return to the last page.\n
   <tt>.up &nbsp;&nbsp;</tt> - Go up one level. (E.g., go from an object to the file containing it.)\n
   <tt>.top &nbsp;</tt> - Show list of files covered by this help.\n
   <tt>.quit&nbsp;</tt> - Leave T3 Help.\n
   <tt>.full&nbsp;</tt> - Show the full page, if only part of it is shown (after a search).\b
   All commands can be abbreviated to their first letter (e.g., <TT>.b</TT>).\b
   <B>Searching</B>\n
   You can type a keyword or part of a keyword, to search. If only one result is found, you'll be taken directly
   to the appropriate page.\n
   If more than one result is found, you'll be asked to pick one by typing <TT>.</TT> (period) followed by the result number.\n
   You can also type <TT>,</TT> (comma) followed by a text to search among the results.\n
   If there's more than 22 results, your only option is to search among results or to do a new search.\b
   To jump to a specific keyword, type <tt>!</tt> followed by the keyword. This search is case sensitive.
   E.g., <tt>!Sense</tt> will take you to the Sense class directly, instead of showing results like sense.t and sense properties.
   <TT>!!Sense</TT> will furthermore show the entire Sense page, instead of only the selected parts.\b
   To do a full-text search (somewhat slow!) type <TT>*</TT> followed by a single word to search for. Then use <TT>*,</TT> to search among results.\n
   You can also combine this with regular keyword search.\b
   <B>Cross-reference words</B>\n
   All keywords that appear in the index, are highlighted in green. Clicking on them is the same as typing <TT>!</TT> followed by the keyword.\n
   The only exception is that the words ";
  for (local i = 1; i <= mainPage.stopWords.length(); i++) {
   if (i != 1) {
    if (i == mainPage.stopWords.length()) " and "; else ", ";
   }
   "<A plain href=\"<< mainPage.stopWords[i] >>\"><FONT color=#008000><< mainPage.stopWords[i] >></FONT></A>";
  }
  " are not normally linked, as they appear pretty much everywhere.\b";
 }
}

function handleResults(results) {
 globals.results = results;
 
 if (dataType(results) == TypeObject && !results.ofKind(Collection)) { results.makeCurrent(globals.searchKeywords); return; }
 if (results.length() == 0) {
  "No results found for";
  for (local i = 1; i <= globals.searchKeywords.length(); i++) " \"<< globals.searchKeywords[i].htmlify() >>\"";
  ".\n";
  return;
 }
 globals.results = results = results.getUnique();
 if (results.length() == 1) { results[1].makeCurrent(globals.searchKeywords); return; }
 if (results.length() > 22) {
  "\bFound << results.length() >> results.\nEnter <TT>,</TT> followed by a keyword or a part of a keyword, to search among results.\n";
  return;
 }

 "\b<B>Pages matching";
 for (local i = 1; i <= globals.searchKeywords.length(); i++) " \"<< globals.searchKeywords[i].htmlify() >>\"";
 "</B><TABLE cellspacing=0 cellpadding=0>";
 for (local i = 1; i <= results.length(); i++) {
  "<TR><TD>\n<B>.<< i >>&nbsp; </B><TD><A plain href=\".<< i >>\"><FONT color=#000080><< results[i].title >></FONT></A> \t<TD>";
  if (results[i].parent) "(<< results[i].parent.title >>)";
 }
 "</TABLE>\n";
}

function subsearch(s) {
 if (globals.searchKeywords == nil) { "No previous search performed.\n"; return; }
 local results = globals.results;
 local newresults = new Vector(50);
 globals.searchKeywords.append(s);
 for (local i = 1, local len = results.length(); i <= len; i++) {
  local obj = results[i];
  for (local j = 1; j <= obj.keywords.length(); j++)
   if (obj.keywords[j].toLower().find(s)) newresults += obj;
 }
 if (newresults.length() == 0) "No results. Enter <tt>,<tt> followed by a search text to search among the old result set.\n";
 else if (newresults.length() == results.length()) "That didn't narrow the results down at all.";
 else handleResults(newresults);
}

function search(s) {
 local results = new Vector(50);
 globals.searchKeywords = [s];
 for (local i = 1, local len = globals.keywordIndex.length(); i <= len; i++)
  if (globals.keywordIndex[i].toLower().find(s)) results += globals.pageIndex[i];
 globals.showFullPage = nil;
 handleResults(results);
}

function absSearch(s) {
 if ((globals.showFullPage = (s.substr(1, 1) == '!')) != nil) s = s.substr(2);
 local results = globals.keywordLookup[s];
 if (!results) "Keyword not found in index. Type it without the <TT>!</TT> prefix to search without case sensitivity. ";
 else {
  globals.searchKeywords = [s];
  handleResults(globals.pageIndex[results]);
 }
}

function freeSearch(s) {
 globals.showFullPage = nil;
 "Free-text search not implemented. ";
}

function freeSubsearch(s) {
 "Free-text search not implemented. ";
}

function goUp() {
 if (globals.curPage.parent) globals.curPage.parent.makeCurrent(nil);
 else "This is the top level page. You can't go up from here.\n";
}

function showFull() {
 if (globals.curPage.partKeyword) {
  globals.curPage.partKeyword = nil;
  globals.curPage.makeCurrent(nil);
 } else "This is the full page.\n"; 
}

function goResult(resNo) {
 if (resNo <= 0 || resNo > globals.results.length()) "Number out of range.\n";
 else globals.results[resNo].makeCurrent(globals.searchKeywords);
}

function main(args) {
 local cmd, cmdLC, cmdprefix;
 globals.banner = bannerCreate(nil, BannerFirst, nil, BannerTypeText, BannerAlignTop, 2, BannerSizeAbsolute, 0);
 
 "<BODY bgcolor=#FFFFFF text=#000000><TITLE>T3 Help</TITLE>";
 mainPage.addPage(instructions);
 mainPage.makeCurrent(nil);
 
 "\bType a keyword (the name of a symbol or library file) to browse the help system.\n
  Type <A plain href=\"?\"><B><FONT color=#000080>&nbsp;?&nbsp;</FONT></B></A> for help on this help system, or <A plain href=\".quit\"><B><FONT color=#000080>.quit</FONT></B></A> to exit.";

 while (true) {
  "\b&gt; "; cmd = inputLine();
  if (!cmd || (cmdLC = cmd.toLower()) == '.quit' || cmdLC == '.q') break;
  if (cmd == '') { "Eh? (Type ? for help.)\n"; continue; }
  if (cmd == '?') instructions.makeCurrent(nil);
  else if ((cmdprefix = cmd.substr(1, 1)) == '.') {
   cmd = cmdLC.substr(2);
   if (cmd is in ('b', 'back'))	globals.lastpage.makeCurrent(globals.lastpage.partKeyword);
   else if (cmd is in ('t', 'top'))	mainPage.makeCurrent(nil);
   else if (cmd is in ('u', 'up'))	goUp();
   else if (cmd is in ('f', 'full')) showFull();
   else if (rexMatch('[0-9]+$', cmd)) goResult(toInteger(cmd));
   else "Unknown command. Type ? for help.";
  } else if (cmdprefix == ',') subsearch(cmdLC.substr(2));
  else if (cmdprefix == '!') absSearch(cmd.substr(2));
  else if (cmdprefix == '*') {
   if (cmd.substr(2, 1) == ',') freeSubsearch(cmdLC.substr(3));
   else freeSubsearch(cmd.substr(2));
  } else search(cmdLC);
 }
 "\b<B>*** You have (hopefully) been helped ***</B>\n";
}
