#charset "us-ascii"

/* Copyright (c) 2003 Sren J. Lvborg.  All Rights Reserved. */

/*
 *   t3helpgen.t - TADS 3 helpfile generator
 */

/////////////// DEFINITIONS ///////////////

#define BASE_PATH '/Program/TADS3/'

#define MAIN_INC_PATH  'include/'
#define MAIN_LIB_PATH  'lib/'
#define ADV3_LIB_PATH  'lib/adv3/'
#define LANG_LIB_PATH  'lib/adv3/en_us/'

#define MAIN_INC_FILES ['array.h', 'bignum.h', 'bytearr.h', 'charset.h', \
   'dict.h', 'file.h', 'gramprod.h', 'lookup.h', 'reflect.h', 'strcomp.h', \
   'systype.h', 't3.h' , 'tads.h', 'tadsgen.h', 'tadsio.h', 'tok.h', \
   'vector.h', 't3test.h']

#define MAIN_LIB_FILES \
   ['_main.t', 'file.t', 'tok.t', 'gameinfo.t', 'reflect.t']

#define ADV3_LIB_FILES ['action.t', 'actions.t', 'actor.t', 'adv3.h', \
   'banner.t', 'disambig.t', 'events.t', 'exec.t', 'exits.t', 'extras.t', \
   'footnote.t', 'hintsys.t', 'input.t', 'lister.t', 'menusys.t', 'misc.t', \
   'modid.t', 'numbers.t', 'objects.t', 'output.t', 'parser.t', 'pov.t', \
   'precond.t', 'report.t', 'resolver.t', 'score.t', 'sense.t', 'status.t', \
   'thing.t', 'travel.t', 'verify.t']

#define LANG_LIB_FILES ['en_us.h', 'en_us.t', 'instruct.t', 'msg_neu.t']

#include <tads.h>
#include <file.h>
#include "t3help_tok.h"

enum modeTopLevel, modePropDeclOrExport, modeEnum;
enum omNormal, omModify, omReplace;

#define TEXT_PLAIN	0x0000	// Just show the text plainly.
#define TEXT_BOLD		0x0001	// Wrap the text in <B> ... </B> tags.
#define TEXT_TAGON 	0x0002	// Text follows which are to be shown in the <A>...</A>, but isn't a part of the keyword.
#define TEXT_KEYWORD	0x0004	// Link the text (<A href="!...">...</A>)
#define TEXT_TELETYPE	0x0008	// Wrap the text in <TT> ... </TT> tags.

/////////////// GAMEINFO ///////////////

PreinitObject
 execute() {
  local tab = new LookupTable();
  tab['Name'] = 'T3 Help Generator';
  tab['Version'] = '1.0';
  tab['Byline'] = 'by Sren J. Lvborg';
  tab['HtmlByline'] = 'by S&oslash; J. L&oslash;vborg';
  tab['AuthorEmail'] = 'Sren Lvborg <your-email@your-address.com>';
  tab['Desc'] = 'Extracts source code comments from the TADS libraries and generates an include file for T3 Help.';
  tab['HtmlDesc'] = tab['Desc'];
  tab['PresentationProfile'] = 'Multimedia';

  writeGameInfo(tab, 'GameInfo.txt');
 }
;

/////////////// GLOBALS ///////////////

globals: object {
 // Symbols covered by the helpfile:
 keywordLookup	= static new LookupTable(512, 4096); // Index: Keyword string. Value: Keyword ID
 keywordIndex	= static new Vector(4096); // Index: Keyword ID. Value: Keyword string.
 pageIndex 		= static new Vector(4096); // Index: Keyword ID. Value: Page object or list of page objects.
 // Don't link these words, unless they have a tagon - "()".
 stopWords		= ['length', 'sublist', 'say', 'append', 'match', 'destination', 'display', 'name', 'location', 'noun', 'material', 'seen'];

// These are just the remains of a very bad compression (read: expansion) scheme. Ignore them.
// nonKeywordLookup = static new LookupTable(512, 4096); // Index: Word. Value: Word ID.
// nonKeywordIndex	= static new Vector(4096); // Index: Word ID. Value: Word.
 
 idents	= static new LookupTable(512, 2048); // Identifiers used in the generated t3help_hlp.h.
 WaitingContent = static new LookupTable(128, 256); // Content like export clauses.
 banner	= nil
 
// session = 1;

/* BAAAAD compression scheme, I tell you...

 getNonKeywordID(nonKeyStr) {
  local x = nonKeywordLookup[nonKeyStr];
  if (!x) {
   nonKeywordIndex.append(nonKeyStr);
   nonKeywordLookup[nonKeyStr] = (x = nonKeywordIndex.length());
  }
  return x;
 }
*/
 
 addKeyword(keyStr, page) {
  local keyID = keywordLookup[keyStr];
  if (!keyID) { // New keyword
   pageIndex.append(page);
   keywordIndex.append(keyStr);
   keywordLookup[keyStr] = (keyID = keywordIndex.length());
   return; // Done.
  }
  local pg = pageIndex[keyID];
  
  if (!pg.ofKind(Vector)) {
   local v = new Vector(5);
   v.append(pg);
   v.append(page);
   pageIndex[keyID] = v;
  }
  else pg.append(page);
 }
 
 addIdent(suggestedname, obj) {
  local i = 1;
 
  for (i = 1; i <= suggestedname.length(); i++) { // Make the name a valid identifer.
   local uni = suggestedname.toUnicode(i);
   if ((uni >= 65 && uni <= 90) || (uni >= 97 && uni <= 122) || uni == 95 ||
       (uni >= 48 && uni <= 57 && i > 1)) continue; // If [A-Za-z_] or ([0-9] and not first ch.)
   suggestedname = suggestedname.substr(1, i - 1) + '_' + suggestedname.substr(i + 1); i--;
  }
  if (idents.isKeyPresent(suggestedname)) { // Identifier in use.
   while (idents.isKeyPresent(suggestedname + i)) i++;
   suggestedname += i;
  }
  idents[suggestedname] = obj;
  obj.ident = suggestedname;
 }

 bannerProgress(txt) {
  bannerClear(banner);
  bannerSay(banner, '<BODY bgcolor=statusbg text=statustext>');
  bannerSay(banner, txt);
  inputEvent(1); // Refresh screen
 }
   
 rexCHr		= static new RexPattern('-+');
 rexCCopy		= static new RexPattern('Copyright (.+)');
 rexCHeader		= static new RexPattern('[a-zA-Z_][a-zA-Z_0-9]*<dot>[th] - (.*)');
 rexFindSymbol	= static new RexPattern('([^a-zA-Z_0-9]*)([a-zA-Z_0-9]+)(<lparen><rparen>)?');

 // We assume that '<' and '"' are illegal in filenames.
 rexInclude		= static new RexPattern('#include<space>+["<](.+?)[">]');
 rexCharset		= static new RexPattern('#charset<space>+"(.+?)"');
 rexIfdef		= static new RexPattern('#ifdef<space>+([a-zA-Z_][a-zA-Z_0-9]*)');
 rexDefine		= static new RexPattern('#define<space>+([a-zA-Z_][a-zA-Z_0-9]*)<space>*(.+)');
 
 modifycnt  = 0;
 anonobjcnt = 0;
}

/////////////// UTILITY FUNCTIONS ///////////////

function quoteStr(s) {
 return s.findReplace('\\',  '\\\\', ReplaceAll).findReplace('\'', '\\\'', ReplaceAll).
          findReplace('\<<', '\\<',  ReplaceAll).findReplace('"',  '\\"',  ReplaceAll).
          findReplace('\n',  '\\n',  ReplaceAll).findReplace('\b', '\\b',  ReplaceAll).
          findReplace('\t',  '\\t',  ReplaceAll);
}

// Formats str and appends it (possibly in several parts) to v. Adds a line break to the end.

function formatOutput(v, str, boldkw, attribs) {
 local x, symbol, tagon, kw_attr, text;
 
 str = Trim(str).htmlify(HtmlifyTranslateWhitespace);
 kw_attr = attribs | TEXT_KEYWORD;
 if (boldkw) kw_attr |= TEXT_BOLD;
 
 text = '';
 
 while ((x = rexSearch(globals.rexFindSymbol, str)) != nil) {
  text += rexGroup(1)[3]; // The text before the symbol.
  symbol = rexGroup(2)[3];
  tagon  = rexGroup(3); if (tagon) tagon = tagon[3]; else tagon = '';

  str = str.substr(x[1] + x[2]);

  x = globals.keywordLookup[symbol]; 
  if (x && (!globals.stopWords.indexOf(symbol) || tagon != '')) {
   if (text != '') {
    if (attribs != TEXT_PLAIN) v.append(attribs);
    v.append(text); text = '';
   }
   if (tagon == '') v.append(kw_attr); else v.append(kw_attr | TEXT_TAGON);
   v.append(x);
   if (tagon != '') v.append(tagon);
  }
  else text += symbol + tagon;
 }
 text += str + '\n';
 if (attribs != TEXT_PLAIN) v.append(attribs);
 v.append(text);
}

function writesubvalue(f, val) {
 local dt = dataType(val); 
 if ((dt == TypeList) || (dt == TypeObject && val.ofKind(Vector))) {
  f.writeFile('[');
  for (local i = 1, local len = val.length(); i <= len; i++) { if (i > 1) f.writeFile(', ');  writesubvalue(f, val[i]); }
  f.writeFile(']');
  return;
 }
 switch (dt) {
  case TypeObject:	f.writeFile(val.ident);				return;
  case TypeSString:	f.writeFile('\'' + quoteStr(val) + '\'');	return;
  case TypeInt:		f.writeFile(toString(val));			return;
//  case TypeNil:		f.writeFile('nil');				return;
  default: "\bCvtValueToStr ERROR!!\b";
 }
}

function writevalue(f, name, val) {
 if (dataType(val) != TypeNil) { f.writeFile(' ' + name + ' = '); writesubvalue(f, val); f.writeFile(';\n'); }
}

/////////////// PAGE AND CONTENT OBJECTS ///////////////

property text;

class cDefinition: object	{ gettext(v) { formatOutput(v, Trim(text),	 nil,  TEXT_TELETYPE);	} }
class cHelptext: object		{ gettext(v) { formatOutput(v, '\n' + Trim(text),true, TEXT_PLAIN); 	} }
class cCode: object		{ gettext(v) { formatOutput(v, '\n' + Trim(text),nil,  TEXT_TELETYPE); 	} }
cHR: object				{ gettext(v) { v.append('<HR>');							} }
cSmallHR: object			{ gettext(v) { v.append('\b'); /* '<HR width=100 align=left>'; */		} }

property keywords, ident, title, parent, pageindex, subpages, contents;

class page: object {
 classname = 'page';
 construct(atitle) {
  title = atitle;
  subpages = new Vector(20);
  contents = new Vector(10);
  keywords = new Vector(10);
  globals.addIdent('pg_' + title, self);    
 }
 
 addPage(pg)	{ subpages += pg;  pg.parent = self;  pg.pageindex = subpages.length(); }
 addHR()		{ contents += cHR; }
 addSmallHR()	{ contents += cSmallHR; }
 addHelp(text)	{ local x = new cHelptext;  x.text = text;  contents += x; }
 addCode(text)	{ local x = new cCode;      x.text = text;  contents += x; }
 
 writefields(f) {
  f.writeFile(' title = \'' + quoteStr(title) + '\';\n');
  writevalue(f, 'parent', parent);
  writevalue(f, 'pageindex', pageindex);
  writevalue(f, 'subpages', subpages);
  writevalue(f, 'keywords', keywords);

  local v = new Vector(100);
  for (local i = 1, local len = contents.length(); i <= len; i++)
   contents[i].gettext(v);
  
  f.writeFile(' text = [');
  local x = 1;
  for (local i = 1, local len = v.length(); i <= len; i++) {
   if (i != 1) {
    if ((x += 8) > 600) { f.writeFile(',\n'); x = 1; }
    else f.writeFile(', ');
   }
   
   local itm = v[i];
   writesubvalue(f, itm);
   if (dataType(itm) == TypeSString) x += itm.length();
  }
  f.writeFile('];\n');
 }
 write(f) {
  f.writeFile(ident + ': ' + classname + ' {\n');
  writefields(f);
  f.writeFile('}\n\n');
 }
 preparewrite() { } // This is called on all pages before any page is actually written.
 forAllPages(func) {
  func(self);
  local len = subpages.length();
  for (local i = 1; i <= len; i++) subpages[i].forAllPages(func);
 }
}

/////////////// SPECIALIZED PAGES ///////////////

class defpage: page { // A definition page.
 classname = 'defpage';
 // keywords stores a list of symbols defined on this page. (Strings.)

 construct(def, symbols, title_) {
  title = title_;
  subpages = new Vector(20);
  contents = new Vector(10);
  keywords = new Vector(10);
  addDefinition(def, symbols);
  globals.addIdent('dp_' + title, self);    
 }
 
 addDefinition(def, symbols) {
  keywords += symbols; // See? asymbols can be either a list of strings, or just a string. Yay!

  local x = new cDefinition;
  x.text = def;
  contents += x;
 }
}

class objpage: defpage {
 classname = 'objpage';
  
 superclasses = [];
 subclasses = [];
 replacedby = [];
 modifiedby = [];
 replaces = nil;
 modifies = nil;
 objectmode = omNormal;

 writefields(f) {
  inherited(f);
  writevalue(f, 'superclasses', superclasses);
  writevalue(f, 'subclasses', subclasses);
  writevalue(f, 'modifiedby', modifiedby);
  writevalue(f, 'replacedby', replacedby);
  writevalue(f, 'modifies', modifies);
  writevalue(f, 'replaces', replaces);
 } 
 preparewrite() {
  for (local j = 1; j <= superclasses.length(); j++) {
   local super = globals.keywordLookup[superclasses[j]];
   if (super)
    (superclasses[j] = globals.pageIndex[super]).subclasses += self;
  }
  if (modifies) {
   local mod = globals.keywordLookup[modifies];
   if (mod)
    (modifies = globals.pageIndex[mod]).modifiedby += self;
  }
  if (replaces) {
   local repl = globals.keywordLookup[replaces];
   if (repl)
    (replaces = globals.pageIndex[repl]).replacedby += self;
  }
 }
}

/////////////// FILE PAGE (WHERE THE REAL WORK IS DONE) ///////////////

property path, includes, tokens, tokensLength;

#define nextToken()		tokenIndex += 3
#define getToken()		tokens[tokenIndex]
#define getTokenValue()		tokens[tokenIndex + 1]
#define getTokenText()		tokens[tokenIndex + 2]
#define getNextToken()		tokens[(tokenIndex += 3)]
#define getNextTokenValue()	tokens[(tokenIndex += 3) + 1]
#define getNextTokenText()	tokens[(tokenIndex += 3) + 2]
#define peekToken()		tokens[tokenIndex + 3]
#define peekTokenValue()	tokens[tokenIndex + 4]

class filepage: page {
 classname = 'filepage';
 copyright = '';
 header = '';
 charset = '';

 miscPage = nil;

 construct(apath, afilename) {
  inherited (afilename);
  path = apath;
  includes = new Vector(10);
  mainPage.addPage(self);
  keywords = [afilename];
 }
  
 tokenIndex = 1;
 errors = 0;

 Error(msg) { // This returns after showing the error.
  if (++errors == 1) "\n<B><< title >></B>"; // First error? Then show filename.
  "\n<< msg >>\n";
 }

 FatalError(msg) { // This never returns. (It throws an exception.)
  if (++errors == 1) "\n<B><< title >></B>"; // First error? Then show filename.
  "\nFatal error: << msg >>\n";
  throw new Exception;
 }
  
 SyntaxError(expected) { // This returns after handling the error.
  if (++errors == 1) "\n<B><< title >></B>"; // First error? Then show filename.
  
  "\nSyntax error: Expected << expected >>, found \"<< RemoveNewlines(getTokenText()) >>\".\n<TT>";
  
  for (local i = tokenIndex - 7; i < tokenIndex; i += 3) "<< RemoveNewlines(tokens[i]) >>";
  " </TT><B>[ ERROR -&gt; ]</B><TT> ";
  local max = tokenIndex + 16;
  if (max > tokensLength) max = tokensLength;
  for (local i = tokenIndex + 2; i < max; i += 3) "<< RemoveNewlines(tokens[i]) >>";
  "</TT>\b
   Attempting to skip to the next top-level statement starting with a reserved word... ";

  while (peekToken() not in (rwClass, rwProperty, rwFunction, rwExport, rwDictionary, rwModify,
   rwReplace, rwIntrinsic, rwEnum, rwTransient)) {
   nextToken();
   if (tokenIndex >= tokensLength - 2) { "But reached the end of file without any luck.\n"; return; }
  }
  "Found <TT><< peekTokenValue() >></TT>.\n";
 }
 
 readFunction(str, id) {
  local bc = 0;
  
  str += getNextTokenText();		// '('

  do str += getNextTokenText();	// args, ')'
  while (getToken() != tokParenE);

  local pg = new defpage(str, id, id); addPage(pg);
  getNewComments(pg);

  do {
   if (getNextToken() == tokBraceB) bc++;
   else if (getToken() == tokBraceE) bc--;
  } while (bc > 0);
  clearNewComments(); // Don't store comments appearing within function code.
 }

 readCommaList(outlist) { // outlist is the vector to store the list in.
  local str = '';
  while (true) {
   outlist.append(getNextTokenValue());		// the list item
   str += getTokenText();

   if (getNextTokenValue() != ',') break;		// the ','
   str += getTokenText();
  }
  return str;
 }

 readTemplate(str, id) {
  do {
   if (tokenIndex >= tokensLength - 2) FatalError('EOF encountered while reading template for class ' + id + '.');
   str += getNextTokenText();
  } while (getToken() != tokSemicolon);
  
  local x = globals.WaitingContent[id];
  if (x) x.addDefinition(str, id);
  else globals.WaitingContent[id] = (x = new defpage(str, id, id));
  getNewComments(x);
 }
 
 readIntrinsic(str) {
  local curTok = getNextToken();
  local symbol, iid, current_page, curTokStr;
  
  if (curTok == rwClass) {
   str += getTokenText();
   symbol = getNextTokenValue();	str += getTokenText();
   iid = getNextTokenText();		str += ' ' + iid;

   local superCl = new Vector(10);
   if (peekToken() != tokBraceB) str += getNextTokenText() + readCommaList(superCl);
   else nextToken();
   current_page = new objpage(str, symbol, symbol);
   current_page.superclasses = superCl;
  } else {
   iid = getTokenText(); str += ' ' + iid;
   current_page = new defpage(str, [], iid);
   current_page.keywords.append(iid);
   nextToken();
  }
  addPage(current_page); getNewComments(current_page);
  if (getToken() != tokBraceB) { SyntaxError('"{"'); return; }

  while ((curTok = getNextToken()) != tokBraceE) {
   if (curTok == rwStatic) { str = getTokenText(); nextToken(); } else str = '';

   local id = getTokenValue();
   str += getTokenText();
      
   while (curTokStr = getNextTokenText(), getToken() != tokSemicolon) str += curTokStr;
   current_page.addSmallHR();
   current_page.addDefinition(str, id);
   getNewComments(current_page);
  }
  getNewComments(current_page);
 }
 
 readObject(str, id, om) {
  local bc = 0;
  local delimiter = tokSemicolon;
  local superCl;

  if (getNextTokenValue() == ':') {
   str += ':';	// ':'
   superCl = new Vector(2);
   str += readCommaList(superCl);
  } else {
   superCl = [id];
   id = nil;
  }
  
  if (getToken() == tokBraceB) { delimiter = tokBraceE; nextToken(); }

  if (!id) {
   if (om == omModify) id = 'modify ' + superCl[1] + ' ' + (++globals.modifycnt);
   else id = 'anonobj ' + (++globals.anonobjcnt);
  }
  if (om == omReplace) { bc = id; id = 'replace ' + id; }
  local current_page = new objpage(str, id, id);  addPage(current_page);  getNewComments(current_page);
  if (om == omReplace) current_page.replaces = bc;
  else if (om == omModify) { current_page.modifies = superCl[1];  superCl = []; }

  current_page.objectmode = om;
  current_page.superclasses = superCl;

  // Now, let's read the properties.

  str = '';
  bc = 1;
  local curTok, curTokStr, propname;
  
  for (; (curTok = getToken()) != delimiter || bc > 1; nextToken()) {
   curTokStr = getTokenText();
   if (bc == 1) {
    if (curTok == tokSymbol) {
     str = curTokStr;
     propname = getTokenValue(); // A property identifier.
    } else if (curTok == tokParenB) {
     str += curTokStr;
     do str += getNextTokenText();
     while (getToken() != tokParenE);
    } else if (curTok == tokBraceB) {
     current_page.addSmallHR;
     current_page.addDefinition(str, propname);
     getNewComments(current_page);
     str = '';
    } else if (curTok == tokAssign) { // '='
     str += curTokStr;
     if (peekTokenValue() == '&') str += getNextTokenText();
     if (peekToken() == rwStatic) str += getNextTokenText();
     if (peekToken() == rwNew) {
      str += getNextTokenText();
      if (peekToken() == rwTransient) str += getNextTokenText();
      str += getNextTokenText(); // Read 'new' ['transient'] <class>.
     }

     local parencnt = 0;
     for (;;) {
      str += getNextTokenText();
      curTokStr = getTokenValue();
      if (curTokStr is in ('(', '[')) parencnt++;
      else if (curTokStr is in (')', ']')) parencnt--;
      if (parencnt == 0) break;
     };
     
     current_page.addSmallHR;
     current_page.addDefinition(str, propname);
     getNewComments(current_page);
     str = '';
    };
   } else if (curTok == tokBraceE) clearNewComments(); // Don't store comments found within methods/prop declarations.
   if (curTok == tokBraceB) bc++;
   else if (curTok == tokBraceE) bc--;

   if (tokenIndex >= tokensLength - 2) FatalError('EOF encountered while reading object ' + id + '.');
  }
  getNewComments(current_page);
 }

 readPreproc(str) {
  do str += getNextTokenText();
  while (getToken() != tokPreprocEnd);

  local i = rexMatch(globals.rexInclude, str);
  if (i) { includes += rexGroup(1)[3]; return; }

  i = rexMatch(globals.rexCharset, str);
  if (i) { charset += rexGroup(1)[3]; return; }

//       i = rexMatch(globals.rexIfdef, curTokStr);
//       if (i) { ifdefs += rexGroup(1)[3]; return; }
// rexInclude		= static new RexPattern('#include<space>+["<](.+?)[">]');
// rexIfdef		= static new RexPattern('#ifdef<space>+([a-zA-Z_][a-zA-Z_0-9]*)');
// rexDefine		= static new RexPattern('#define<space>+([a-zA-Z_][a-zA-Z_0-9]*)<space>*(.+)');
//       if 

  self.addCode(str);
  getNewComments(self);
 }
 
 // These two methods keeps track of the comments that has appeared since the last check.
 lastCommentCheck = 1;
 getNewComments(pg) {
  local len = comments.length();
  local cmts = new Vector(5);
  while (lastCommentCheck < len && comments[lastCommentCheck] < tokenIndex) cmts.append(comments[(lastCommentCheck += 2) - 1]);
  len = cmts.length();
  if (len == 0) return;
  
  local contlen = pg.contents.length();
  local latestpgcont;
  if (contlen > 0) {
   latestpgcont = pg.contents[contlen];
   pg.contents.setLength(contlen - 1);
  }
  
  for (local i = 1; i <= len; i++) {
   local cmt = cmts[i];
   if (rexMatch(globals.rexCHr, cmt)) {
    if (pg.contents.length() > 1) pg.addHR;
   } else if (rexMatch(globals.rexCCopy, cmt)) copyright += rexGroup(1)[3];
   else if (rexMatch(globals.rexCHeader, cmt)) header += rexGroup(1)[3];
   else pg.addHelp(cmt);
  }
  if (latestpgcont) {
   contlen = pg.contents.length();
   if (contlen > 0) pg.contents.insertAt(contlen, latestpgcont); // Insert the (old) last page content just before the last cmt.
   else pg.contents.append(latestpgcont);
  }
 }
 clearNewComments() {
  local len = comments.length();
  while (lastCommentCheck < len && comments[lastCommentCheck] < tokenIndex) lastCommentCheck += 2;
 }
 
 bannerProgress(txt) {
  if (!globals.banner) return;
  bannerClear(globals.banner);
  bannerSay(globals.banner, '<BODY bgcolor=statusbg text=statustext><B>' + title + '</B>: ');
  bannerSay(globals.banner, txt);
  inputEvent(1); // Refresh screen
 }
 
 parse() {
  bannerProgress('Reading/pretokenizing');
  tadsTokenizer.parent = self; // We'll handle the errors.

  local f = File.openTextFile(path + title, FileAccessRead); // title stores filename.
  local str;
  tadsTokenizer.ptoks	= new Vector(2560);
  tadsTokenizer.str	= '';

  local x = 0;
  local treshold = 4096; // We try reading "only" 4K at a time.
  
  while ((str = f.readFile()) != nil) {
   tadsTokenizer.str += str;
   if ((x++ & 0xFF) == 0)  bannerProgress('Reading/pretokenizing.' + makeString('.', x/256));
   if (tadsTokenizer.str.length() > treshold) {
    if (!tadsTokenizer.pretokenize()) treshold <<= 2;  // Not enough? Try reading four times as much.
    else treshold >>= 1;  // Happy with this amount? Try reading only half as much next time.
   }
  }

  while (tadsTokenizer.str.length() != 0)
   if (!tadsTokenizer.pretokenize())
    FatalError('Pretokenizer error: Unterminated string/comment.');

  f.closeFile();
  
  bannerProgress('Tokenizing/analyzing syntax.');
  tadsTokenizer.tokenize();
//  for (local i = 3; i <= tadsTokenizer.toks.length(); i += 3) "<< tadsTokenizer.toks[i].htmlify() >> <FONT color=#808080> ~ </FONT>";
  
  tokens = tadsTokenizer.toks;
  tokensLength = tokens.length();
  comments = tadsTokenizer.comments;
  tadsTokenizer.toks = nil; // Make sure we don't keep this load of data around after we're done using it.
  tadsTokenizer.comments = nil;

  local id, curTok, curTokStr, curTokValue;

  for (tokenIndex = 1; tokenIndex < tokensLength; nextToken()) {
   curTok = getToken(); curTokStr = getTokenText(); curTokValue = getTokenValue();

   switch (curTok) {
    case tokPreproc: readPreproc(curTokStr);  break;

     // An identifier
    case tokSymbol:
     handlesymbol:
     id = peekToken();
     if (id == tokParenB) readFunction(curTokStr, curTokValue);
     else if (id == rwTemplate) readTemplate(curTokStr, curTokValue);
     else readObject(curTokStr, curTokValue, omNormal);
    break;

    /// Reserved words ///
	
    case rwTransient:
     curTokStr += getNextTokenText();
     curTok = getToken(); curTokValue = getTokenValue();
     if (curTok == tokSymbol) goto handlesymbol;
     if (curTok != rwClass) { SyntaxError('"class" or symbol (following transient)'); break; }
          
    case rwFunction:
    case rwClass:
    case rwModify:
    case rwReplace:
     id = getNextTokenValue();
     curTokStr += getTokenText();

     if (curTok == rwFunction) { readFunction(curTokStr, id); break; }
     if (curTok == rwReplace) {
      if (peekToken() == tokParenB) readFunction(curTokStr, id);
      else readObject(curTokStr, id, omReplace);
      break;
     }
     readObject(curTokStr, id, (curTok == rwModify) ? omModify : omNormal);
    break;
       
    case rwExport:
     str = curTokStr + getNextTokenText();
     id = getTokenValue();
     str += getNextTokenText();
     if (getToken() == tokSString) str += getNextTokenText();
     if (getToken() != tokSemicolon) { SyntaxError('";"'); return; }

     local x = globals.WaitingContent[id];
     if (x) x.addDefinition(str, id);
     else globals.WaitingContent[id] = (x = new defpage(str, id, id));
     getNewComments(x);
    break;
        
    case rwIntrinsic: readIntrinsic(curTokStr); break;
        
    case rwDictionary:
     curTokStr += getNextTokenText();
     if (getToken() != rwProperty) {
      id = getTokenValue();
      self.addDefinition(curTokStr + getNextTokenText(), id); // Read the ';'
      getNewComments(self);
      break;     
     }
     // A dictionary property.
     
    case rwProperty:
    case rwEnum:
     id = new Vector(1);
     str = curTokStr;
     if (curTok == rwEnum && peekToken() == rwToken) { str += getNextTokenText(); }          
     
     do {
      id.append(getNextTokenValue());
      str += getTokenText();
     } while (getToken() != tokSemicolon);

     self.addDefinition(str, id);
     getNewComments(self);
    break;

    case tokOperator:
     if (curTokValue == '+') {
      if (getNextToken() == rwProperty) { // A "+ property <symbol>;" definition.
       // We, uh, ignore this completely.
       nextToken(); nextToken(); // Skip symbol and ';'
       break;
      }
      SyntaxError('"property"');
      break;
     }
     // Not a '+'. Pass it onto the invalid tokens.

    case tokSemicolon: /* ';' is a valid top-level statement. */ break;
    
    /// Invalid tokens ///
/*
    case tokPunct:
    case tokNumber:
    case tokSString:
    case tokDString: */
    default:
     SyntaxError('definition');
   }
  } // For loop
 }
 
 writefields(f) {
  globals.bannerProgress('Writing <B>' + title + '</B> to t3help_hlp.h.');
  inherited(f);
  writevalue(f, 'copyright', copyright);
  writevalue(f, 'header', header);
  writevalue(f, 'charset', charset);
  writevalue(f, 'includes', includes);
 }
 preparewrite() {
  for (local j = 1; j <= includes.length(); j++) {
   local incfile = globals.keywordLookup[includes[j]];
   if (incfile) includes[j] = globals.pageIndex[incfile];
  }
 }
}

/////////////// THE REMAINING STUFF (ENTRYPOINT, ETC.) ///////////////

mainPage: page {
 construct() {
  subpages = new Vector(20);
  keywords = [];
  globals.addIdent('mainPage', self);    
 }
 write(f) {
  f.writeFile('modify mainPage {\n');
  writevalue(f, 'gendate', getGameInfoToday());
  writevalue(f, 'subpages', subpages);
  writevalue(f, 'stopWords', globals.stopWords);
  f.writeFile('}\n\n');
 }
}

parsefile(p, f) {
 local pg = new filepage(p, f);
 try pg.parse();
 catch (RuntimeError E) "\nParse failed (<< E.exceptionMessage >>).\b";
 catch (Exception E) "\nParse failed.\n<< E.displayException() >>\b";
 pg.tokens = nil; // Don't keep those data around.
 pg.comments = nil;
}

main(args) {
 mainPage.construct();
 globals.banner = bannerCreate(nil, BannerFirst, nil, BannerTypeText, BannerAlignBottom, 2, BannerSizeAbsolute, 0);
 
 "<TITLE>T3 Help Generator - reading library files</TITLE>
  This T3 program will read all the TADS/ADV3 library source files and generate an
  include file (t3help_hlp.h) with the collected data. When finished, compile t3help.t.\b
  Anyway, this is a fairly complex process involving a lot of data, and the TADS interpreter
  may (well, probably will) freeze. This doesn't mean that it has actually crashed, just that
  it's not wasting CPU time refreshing the display and such.\n
  The whole processing will take up to 5 minutes, less if you have just a reasonably fast
  computer, more if it's slow and you do other things while T3 Help Generator is running.\n
  Remember, the library files are 2.6 MB &emdash; there's a lot of text to process.\b
  Using the HTML TADS interpreter, you can safely ignore any \"more\"-prompts that may show up. They don't stop processing,
  only the text scrolling. On other interpreters, this may be different.";
 
 local startTime = getTime(GetTimeTicks);
 
// parsefile(BASE_PATH + ADV3_LIB_PATH, 'travel.t');
// parsefile(BASE_PATH + MAIN_INC_PATH, 't3.h');
// parsefile(BASE_PATH + MAIN_LIB_PATH, '_main.t');

/* --- */

 foreach (local x in MAIN_INC_FILES) parsefile(BASE_PATH + MAIN_INC_PATH, x);
 foreach (local x in MAIN_LIB_FILES) parsefile(BASE_PATH + MAIN_LIB_PATH, x);
 foreach (local x in ADV3_LIB_FILES) parsefile(BASE_PATH + ADV3_LIB_PATH, x);
 foreach (local x in LANG_LIB_FILES) parsefile(BASE_PATH + LANG_LIB_PATH, x);

/* --- */

 "<TITLE>T3 Help Generator - processing data</TITLE>";
 globals.bannerProgress('Generating indexes and translating symbols.');

 mainPage.forAllPages(
  new function(obj) {
   for (local i = 1; i <= obj.keywords.length(); i++) {
    globals.addKeyword(obj.keywords[i], obj);
   }
  } );
 
 globals.WaitingContent.forEachAssoc(
  new function(id, pg) {
   local x = globals.keywordLookup[id];
   if (!x) "\nWarning: << pg.title >> has no page to attach to.\n";
   else {
    x = globals.pageIndex[x];
    if (x.ofKind(Vector)) x = x[1]; // Just "attach" it to an arbitrary page. Not good, I know, but.. eh.
    x.addSmallHR();
    x.contents += pg.contents;
   }
  } );
 
 mainPage.forAllPages({obj: obj.preparewrite()});

 local f = File.openTextFile('t3help_hlp.h', FileAccessWrite);

 "<TITLE>T3 Help Generator - writing helpfile</TITLE>";
 f.writeFile('/*\n * t3help_hlp.h - A mechanically generated help file.\n * Compile t3help.t to browse it.\n */\n\n');

 mainPage.forAllPages({obj: obj.write(f)});

 f.writeFile('globals: object {\n keywordIndex = [\'');
 for (local i = 1, local len = globals.keywordIndex.length(); i <= len; i++) {
  if (i != 1) f.writeFile('\',\n\'');
  f.writeFile(quoteStr(globals.keywordIndex[i]));
 }
 f.writeFile('\'];\n pageIndex = [');
 for (local i = 1, local len = globals.pageIndex.length(); i <= len; i++) {
  if (i != 1) f.writeFile(',\n');
  writesubvalue(f, globals.pageIndex[i]);
 }
 f.writeFile('];\n}\n\n');
 
 f.closeFile();

 local stopTime = getTime(GetTimeTicks);
 local secs = (stopTime - startTime)/1000;

 if (globals.banner) {
  bannerClear(globals.banner); bannerSay(globals.banner, '<CENTER><B>*** You have made a helpfile ***</B></CENTER>');
 }
 "<TITLE>T3 Help Generator - done</TITLE>
  <B><< secs/60 >>:";
  if (secs % 60 < 10) "0";
  "<< secs % 60 >> elapsed, << globals.keywordIndex.length() >></B> keywords in index.\b
  Press Enter to quit.\b&gt;";
 inputLine();
}

/////////////// THE END ///////////////
